import 'dart:convert';
import 'dart:typed_data';

import 'package:fis_common/logger/logger.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
import 'package:web_socket_channel/status.dart' as wsStatus;

import 'handler.dart';

export 'handler.dart';

/// Listener for listening remote notification.
class JsonRpcNotificationListener {
  JsonRpcNotificationListener({
    required this.host,
    required this.port,
    this.protocol = "ws",
    this.token,
  });

  /// Web Socket protocol. Default value is ['ws'].
  final String protocol;

  /// Web Socket host.
  final String host;

  /// Web Socket port.
  final int port;

  /// A sign for relating to the target communication. It's [nullable].
  final String? token;

  WebSocketChannel? _channel;

  final _handlersMap =
      <NotificationTypeEnum, Set<JsonRpcNotificationHandlerBase>>{};

  int _retryCount = 0;
  static const _retryLimit = 5;

  /// Registered handler intance list.
  List<JsonRpcNotificationHandlerBase> get handlers =>
      _handlersMap.entries.expand((e) => e.value).toList();

  /// Status of running.
  bool get running => _channel != null && _channel!.closeCode == null;

  /// Start listen remote notifications.
  void run() {
    _channel = WebSocketChannel.connect(Uri.parse(_buildConnectUrl()));
    _channel!.stream.listen(
      (data) {
        if (!running) return; // diable inner 5s's retention.

        final message = _parseMessageData(data);
        if (message != null) {
          _handleNotification(message);
        }
      },
      onError: (error) {
        logger.e("[JsonRpcListener] websocket errror.", error);
        _retry(error);
      },
    );
    logger.i("[JsonRpcListener] ran.");
  }

  /// Add notification-handler.
  ///
  /// [handler] A notification-handler instance extends class [JsonRpcNotificationHandlerBase].
  void addHandler<T extends JsonRpcNotificationHandlerBase>(
    T handler,
  ) {
    final type = handler.type;
    if (!_handlersMap.containsKey(type)) {
      _handlersMap[type] = {};
    }
    final set = _handlersMap[type]!;
    set.add(handler);
  }

  /// Remove target notification-handler.
  ///
  /// [type] The type of message for targeting handlers.
  void removeHandler<T extends JsonRpcNotificationHandlerBase>(
      NotificationTypeEnum type) {
    if (!_handlersMap.containsKey(type)) return;

    final set = _handlersMap[type]!;
    set.removeWhere((e) => e is T);
  }

  /// Close and dispose connection/other resources.
  Future<void> close() async {
    if (_channel != null) {
      if (_channel!.closeCode == null) {
        await _channel!.sink.close(wsStatus.normalClosure);
      }
      _channel = null;
    }
  }

  void _retry(dynamic error) {
    // https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent/code
    _channel?.sink.close(1013); // code: try again later.
    _channel = null;

    if (_retryLimit > _retryCount) {
      Future.delayed(const Duration(seconds: 1), () {
        _retryCount++;
        run();
      });
    }
  }

  String _buildConnectUrl() {
    final originUrl = "$protocol://$host:$port";
    final url =
        token == null || token!.isEmpty ? originUrl : "$originUrl?token=$token";
    return url;
  }

  NotifyMessage? _parseMessageData(dynamic data) {
    try {
      final uint8Array = Uint8List.fromList(data);
      final byteData = uint8Array.buffer.asByteData();
      final messageType = byteData.getUint8(0);
      final notifyType = byteData.getUint8(1);
      final messageLength = byteData.getUint8(2);
      final messageConentList =
          Uint8List.view(uint8Array.buffer, 6, messageLength);
      final messageConent = Uint8List.fromList(messageConentList);
      final messageText = const Utf8Decoder().convert(messageConent);

      return NotifyMessage(
        length: messageLength,
        messageType: MessageTypeEnum.values[messageType],
        notificationType: NotificationTypeEnum.values[notifyType],
        message: messageText,
      );
    } catch (e) {
      logger.e("[JsonRpcListener] parse message data error.", e);
    }
    return null;
  }

  void _handleNotification(NotifyMessage message) async {
    final type = message.notificationType;
    if (!_handlersMap.containsKey(type)) return;

    final set = _handlersMap[type]!;
    final orderedHandlers = set.toList()
      ..sort((a, b) => a.order.compareTo(b.order));

    for (var handler in orderedHandlers) {
      await handler.handle(message.message);
    }
  }
}
